{-# LANGUAGE QuasiQuotes #-}
{-# LANGUAGE TemplateHaskell #-}
{-# OPTIONS_GHC -Wno-prepositive-qualified-module #-}

module Glest.Egl (
  initializeEgl
) where

import Control.Exception
import Data.Foldable (elem)
import Data.Set (Set)
import qualified Data.Set as Set
import Data.Vector.Storable.Mutable qualified as V
import Foreign
import Foreign.C.Error
import Foreign.C.String
import Glest.Egl.Types
import Glest.Utils.InlineC
import Language.C.Inline qualified as C
import Language.C.Inline.Unsafe qualified as CU
import Quasar.Prelude
import System.Posix.Types (Fd(Fd))
import System.Posix.ByteString.FilePath (RawFilePath, withFilePath)

C.context ctx

C.include "<assert.h>"
C.include "<fcntl.h>"
C.include "<stdint.h>"
C.include "<unistd.h>"
C.include "<EGL/egl.h>"
C.include "<EGL/eglext.h>"

C.verbatim "PFNEGLQUERYDEVICESEXTPROC eglQueryDevicesEXT;"
C.verbatim "PFNEGLQUERYDEVICESTRINGEXTPROC eglQueryDeviceStringEXT;"

data EglException = EglException EGLint
  deriving (Eq, Exception)

instance Show EglException where
  show (EglException 0x3000) = "EGL_SUCCESS"
  show (EglException 0x3001) = "EGL_NOT_INITIALIZED"
  show (EglException 0x3002) = "EGL_BAD_ACCESS"
  show (EglException 0x3003) = "EGL_BAD_ALLOC"
  show (EglException 0x3004) = "EGL_BAD_ATTRIBUTE"
  show (EglException 0x3005) = "EGL_BAD_CONTEXT"
  show (EglException 0x3006) = "EGL_BAD_CONFIG"
  show (EglException 0x3007) = "EGL_BAD_CURRENT_SURFACE"
  show (EglException 0x3008) = "EGL_BAD_DISPLAY"
  show (EglException 0x3009) = "EGL_BAD_SURFACE"
  show (EglException 0x300a) = "EGL_BAD_MATCH"
  show (EglException 0x300b) = "EGL_BAD_PARAMETER"
  show (EglException 0x300c) = "EGL_BAD_NATIVE_PIXMAP"
  show (EglException 0x300d) = "EGL_BAD_NATIVE_WINDOW"
  show (EglException 0x300e) = "EGL_CONTEXT_LOST"
  show _ = "Invalid EGL error number"

loadEglExtensionFunctions :: IO ()
loadEglExtensionFunctions =
  [CU.block|
    void {
      eglQueryDevicesEXT = (PFNEGLQUERYDEVICESEXTPROC)eglGetProcAddress("eglQueryDevicesEXT");
      eglQueryDeviceStringEXT = (PFNEGLQUERYDEVICESTRINGEXTPROC)eglGetProcAddress("eglQueryDeviceStringEXT");
      // if (eglQueryDevicesEXT == NULL) {
      //   return NULL;
      // }
    }
  |]

eglGetError :: IO EglException
eglGetError = EglException <$> [CU.exp| EGLint { eglGetError() } |]

initializeEgl :: IO ()
initializeEgl = do
  loadEglExtensionFunctions

  deviceCount <- throwErrnoIfMinus1 "eglQueryDevicesEXT"
    [CU.block|
      EGLint {
        EGLint count;
        EGLBoolean result = eglQueryDevicesEXT(0, NULL, &count);
        if (!result) {
          return -1;
        }
        return count;
      }
    |]
  traceIO $ "Available EGL devices: " <> show deviceCount

  devices :: V.IOVector EGLDeviceEXT <- V.new (fromIntegral deviceCount)
  V.unsafeWith devices \devicesPtr ->
    throwErrnoIf_ (== 0) "eglQueryDevicesEXT"
      [CU.block|
        EGLBoolean {
          EGLint count;
          EGLBoolean result = eglQueryDevicesEXT($(EGLint deviceCount), $(EGLDeviceEXT* devicesPtr), &count);
          if (count < $(EGLint deviceCount)) {
            return EGL_FALSE;
          }
          return result;
        }
      |]

  V.iforM_ devices \i device -> do
    vendor <- eglQueryDeviceString device [CU.pure|EGLint { EGL_EXTENSIONS }|]
    traceIO $ mconcat ["Device ", show i, ": ", vendor]
    renderNode <- eglTryQueryDeviceString device [CU.pure|EGLint { EGL_DRM_RENDER_NODE_FILE_EXT }|]
    traceIO $ mconcat ["Device ", show i, ": ", show renderNode]

  traceIO "Using device 0"
  device <- V.read devices 0

  display <- throwErrnoIfNull "eglGetPlatformDisplay"
    [CU.exp|EGLDisplay { eglGetPlatformDisplay(EGL_PLATFORM_DEVICE_EXT, $(EGLDeviceEXT device), NULL) }|]

  (major :: EGLint, minor :: EGLint) <- C.withPtrs_ \(majorPtr, minorPtr) ->
    throwErrnoIf_ (== 0) "eglInitialize"
      [CU.exp|EGLBoolean { eglInitialize($(EGLDisplay display), $(EGLint* majorPtr), $(EGLint* minorPtr)) }|]

  traceIO $ mconcat ["EGL ", show major, ".", show minor, " initialized"]

  when (major == 1 && minor < 4) $ fail "Insufficient EGL version: EGL 1.4 is required"

  traceIO . ("EGL version: " <>) =<< eglQueryString display [CU.pure|EGLint { EGL_VERSION }|]
  traceIO . ("EGL vendor: " <>) =<< eglQueryString display [CU.pure|EGLint { EGL_VENDOR }|]
  traceIO . ("EGL client apis: " <>) =<< eglQueryString display [CU.pure|EGLint { EGL_CLIENT_APIS }|]

  eglExtensionString <- eglQueryString display [CU.pure|EGLint { EGL_EXTENSIONS }|]
  traceIO $ mconcat ["EGL extensions: ", eglExtensionString]

  let
    eglExtensions = Set.fromList (words eglExtensionString)
    requiredEglExtensions = Set.fromList [
      "EGL_KHR_no_config_context",
      "EGL_MESA_image_dma_buf_export",
      "EGL_EXT_image_dma_buf_import",
      "EGL_EXT_image_dma_buf_import_modifiers"
      ]
    missingEglExtensions = Set.difference requiredEglExtensions eglExtensions

  unless (Set.null missingEglExtensions) $
    fail $ "Missing EGL extensions: " <> intercalate " " missingEglExtensions

  throwErrnoIf_ (== 0) "eglBindAPI"
    [CU.exp| EGLBoolean { eglBindAPI(EGL_OPENGL_ES_API) } |]

  -- Requires EGL_KHR_no_config_context
  context <- throwErrnoIfNull "eglCreateContext"
    [CU.block|
      EGLContext {
        static const EGLint attributes[] = {
          EGL_CONTEXT_MAJOR_VERSION, 2,
          // terminate list
          EGL_NONE
        };
        return eglCreateContext($(EGLDisplay display), NULL, EGL_NO_CONTEXT, attributes);
      }
    |]

  throwErrnoIf_ (== 0) "eglMakeCurrent"
    [CU.exp|EGLBoolean { eglMakeCurrent($(EGLDisplay display), EGL_NO_SURFACE, EGL_NO_SURFACE, $(EGLContext context)) }|]


eglQueryString :: EGLDisplay -> EGLint -> IO String
eglQueryString display name =
  peekCString =<< throwErrnoIfNull "eglQueryString"
    [CU.exp|char const * { eglQueryString($(EGLDisplay display), $(EGLint name)) }|]

eglQueryDeviceString :: EGLDeviceEXT -> EGLint -> IO String
eglQueryDeviceString device name =
  peekCString =<< throwErrnoIfNull "eglQueryDeviceStringEXT"
    [CU.exp|char const * { eglQueryDeviceStringEXT($(EGLDeviceEXT device), $(EGLint name)) }|]

eglTryQueryDeviceString :: EGLDeviceEXT -> EGLint -> IO (Maybe String)
eglTryQueryDeviceString device name = do
  cstr <- [CU.exp|char const * { eglQueryDeviceStringEXT($(EGLDeviceEXT device), $(EGLint name)) }|]
  if cstr /= nullPtr
    then Just <$> peekCString cstr
    else pure Nothing

splitExtensions :: String -> Set String
splitExtensions = Set.fromList . words

requireExtension :: Set String -> String -> IO ()
requireExtension exts required = unless (elem required exts) $ fail ("Missing extension " <> required)
